/**
 * 
 * AiCiA - Android IRC Client
 * 
 * Copyright (C)2010 GORRY.
 * 
 */

package net.gorry.aicia;

import java.io.IOException;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.schwering.irc.lib.IRCConnection;
import org.schwering.irc.lib.IRCEventListener;
import org.schwering.irc.lib.IRCModeParser;
import org.schwering.irc.lib.IRCUser;
import org.schwering.irc.lib.ssl.SSLIRCConnection;
import org.schwering.irc.lib.ssl.SSLTrustManager;

import android.content.Context;
import android.os.Handler;
import android.os.Message;
import android.os.SystemClock;
import android.text.SpannableStringBuilder;
import android.text.Spanned;
import android.text.style.ForegroundColorSpan;
import android.text.util.Linkify;
import android.util.Log;

/**
 * 
 * IRCサーバ接続処理
 * 
 * @author GORRY
 *
 */
public class IRCServer {
	private static final String TAG = "IRCServer";
	private static final boolean V = false;
	private static final boolean D = false;

	private IRCServerEventListener[] listeners = new IRCServerEventListener[0];
	private final Context me;

	private IRCConnection mConn;
	private final IRCServerConfig mConfig;
	private final ArrayList<ArrayList<SpannableStringBuilder>> mSpanChannelLog;
	private final ArrayList<String> mChannelName = new ArrayList<String>();
	private final HashMap<String, Integer> mChannelNameMap = new HashMap<String, Integer>();
	private final ArrayList<Boolean> mChannelUpdated = new ArrayList<Boolean>();
	private final ArrayList<String> mChannelTopic = new ArrayList<String>();
	private final ArrayList<ArrayList<String>> mChannelUserList = new ArrayList<ArrayList<String>>();
	private String mNowNick = null;
	private String mCurrentChannel;
	private long mNextKeepAliveTime = 0;
	private long mNextReconnectTime = 0;
	private boolean mNeedReconnect = false;
	private final boolean mConnect;
	private String[] mAlertKeywords = null;

	/**
	 * コンストラクタ
	 * @param context コンテキスト
	 * @param connect IRCサーバに接続する実体として使用するときはtrue
	 */
	IRCServer(final Context context, final boolean connect) {
		if (V) Log.v(TAG, "IRCServer()");
		me = context;
		mConnect = connect;
		mConfig = new IRCServerConfig(me);
		createServer();
		mSpanChannelLog = new ArrayList<ArrayList<SpannableStringBuilder>>();
		createChannel(IRCMsg.sSystemChannelName);
	}

	/**
	 * リスナ登録処理
	 * @param l リスナ
	 */
	public synchronized void addEventListener(final IRCServerEventListener l) {
		if (V) Log.v(TAG, "addEventListener()");
		if (l == null) {
			throw new IllegalArgumentException("Listener is null.");
		}
		final int len = listeners.length;
		final IRCServerEventListener[] oldListeners = listeners;
		listeners = new IRCServerEventListener[len + 1];
		System.arraycopy(oldListeners, 0, listeners, 0, len);
		listeners[len] = l;
	}

	/**
	 * リスナ削除処理
	 * @param l リスナ
	 * @return true 正常に削除された
	 */
	public synchronized boolean removeEventListener(final IRCServerEventListener l) {
		if (V) Log.v(TAG, "removeEventListener()");
		if (l == null) {
			return false;
		}
		int index = -1;
		for (int i=0; i<listeners.length; i++) {
			if (listeners[i].equals(l)) {
				index = i;
				break;
			}
		}
		if (index == -1) {
			return false;
		}
		listeners[index] = null;
		final int len = listeners.length - 1;
		final IRCServerEventListener[] newListeners = new IRCServerEventListener[len];
		for (int i = 0, j = 0; i < len; j++) {
			if (listeners[j] != null) {
				newListeners[i++] = listeners[j];
			}
		}
		listeners = newListeners;
		return true;
	}

	/**
	 * サーバ名取得
	 * @return サーバ名
	 */
	public String getServerName() {
		if (V) Log.v(TAG, "getServerName()");
		return mConfig.mServerName;
	}

	/**
	 * サーバ設定のフェッチ
	 * @return サーバ設定
	 */
	public IRCServerConfig fetchConfig() {
		if (V) Log.v(TAG, "fetchConfig()");
		return mConfig;
	}

	/**
	 * サーバ設定の読み込み
	 * @param configId サーバ設定ID
	 */
	public void loadConfig(final int configId) {
		if (V) Log.v(TAG, "loadConfig()");
		disconnect();
		mConfig.loadConfig(configId);
		mNowNick = mConfig.mNick;
		mAlertKeywords = null;
		if ((mConfig.mAlertKeywords != null) && (mConfig.mAlertKeywords.length() > 0) ) {
			mAlertKeywords = mConfig.mAlertKeywords.split("[ ]+");
			if ( mAlertKeywords[0].length() == 0 ) {
				mAlertKeywords = null;
			} else {
				for (int i=0; i<mAlertKeywords.length; i++) {
					mAlertKeywords[i] = mAlertKeywords[i].toUpperCase();
				}
			}
		}
	}

	/**
	 * サーバ設定の保存
	 * @param configId サーバ設定ID
	 */
	public void saveConfig(final int configId) {
		if (V) Log.v(TAG, "saveConfig()");
		mConfig.saveConfig(configId);
	}

	/**
	 * サーバ設定の削除
	 * @param configId サーバ設定ID
	 */
	public void deleteConfig(final int configId) {
		if (V) Log.v(TAG, "deleteConfig()");
		disconnect();
		mConfig.deleteConfig(configId);
	}

	/**
	 * IRClibのクローズ
	 */
	public void close() {
		if (V) Log.v(TAG, "close()");
		if (!mConnect) return;
		disconnect();
		if (mConn != null) {
			mConn.close();
			mConn = null;
		}
		mNowNick = null;
	}

	/**
	 * IRCサーバへの切断発行
	 */
	public void disconnect() {
		if (V) Log.v(TAG, "disconnect()");
		if (!mConnect) return;
		mNeedReconnect = false;
		clearNextKeepAliveTime();
		clearNextReconnectTime();
		stopTimer();
		for (int i=0; i<10; i++) {
			if (!isConnected()) break;
			mConn.doQuit();
			for (int j=0; j<5; j++) {
				if (!isConnected()) break;
				try {
					Thread.sleep(1000);
				} catch (final InterruptedException e) {
					//
				}
			}
		}
		mNowNick = null;
	}

	/**
	 * IRCサーバへの接続発行
	 */
	public void connect() {
		if (V) Log.v(TAG, "connect()");
		if (!mConnect) return;
		mNeedReconnect = true;
		if (mConn != null) {
			mConn.close();
			mConn = null;
		}
		if (mNowNick == null) {
			mNowNick = new String(mConfig.mNick);
		}
		if (mConfig.mUseSsl) {
			mConn = new SSLIRCConnection(
					mConfig.mHost,
					mConfig.mPort,
					mConfig.mPort,
					mConfig.mPass,
					mNowNick,
					mConfig.mUsername,
					mConfig.mRealname
			);
			((SSLIRCConnection)mConn).addTrustManager(new TrustManager());

		} else {
			mConn = new IRCConnection(
					mConfig.mHost,
					mConfig.mPort,
					mConfig.mPort,
					mConfig.mPass,
					mNowNick,
					mConfig.mUsername,
					mConfig.mRealname
			);
		}
		mConn.addIRCEventListener(new IRClibListener());
		mConn.setEncoding(mConfig.mEncoding);
		mConn.setPong(false);
		mConn.setDaemon(true);
		mConn.setColors(false);
		mConn.setTimeout(0);
		try {
			receiveSystemChannel("Connecting to " + mConfig.mHost + ":" + mConfig.mPort + "...", true);
			mConn.connect();
			startTimer();
		} catch (final IOException ex) {
			receiveSystemChannel("Connection error: " + ex.getClass().getName() + ": " + ex.getMessage(), true);
			ex.printStackTrace();
		}
	}

	/**
	 * 自動接続がONならIRCサーバへの接続発行
	 */
	public void connectAuto() {
		if (V) Log.v(TAG, "connectAuto()");
		if (mConfig.mAutoConnect) {
			if (!isConnected()) {
				connect();
			}
		}
	}

	/**
	 * サーバオブジェクトの作成
	 */
	public void createServer() {
		if (V) Log.v(TAG, "createServer()");
		final String dateMsg = IRCMsg.getDateMsg();
		for (int i=listeners.length-1; i>=0; i--) {
			listeners[i].createServer(mConfig.mServerName, dateMsg);
		}
	}

	/**
	 * サーバオブジェクトの削除
	 */
	public void removeServer() {
		if (V) Log.v(TAG, "removeServer()");
		final String dateMsg = IRCMsg.getDateMsg();
		for (int i=listeners.length-1; i>=0; i--) {
			listeners[i].removeServer(mConfig.mServerName, dateMsg);
		}
	}

	/**
	 * チャンネル名からチャンネルIDを得る
	 * @param channel チャンネル名
	 * @return チャンネルID
	 */
	public synchronized int getChannelId(final String channel) {
		if (V) Log.v(TAG, "getChannelId()");
		if (channel == null) return -1;
		final Object obj = mChannelNameMap.get(channel.toLowerCase());
		if (obj == null) {
			return -1;
		}
		return ((Integer)obj).intValue();
	}

	/**
	 * チャンネルオブジェクトの作成
	 * @param channel チャンネル名
	 */
	public synchronized void createChannel(final String channel) {
		if (V) Log.v(TAG, "createChannel()");
		final int id = getChannelId(channel);
		if (id < 0) {
			final String dateMsg = IRCMsg.getDateMsg();
			final int num = mChannelName.size();
			mSpanChannelLog.add(new ArrayList<SpannableStringBuilder>());
			mChannelName.add(channel);
			mChannelUpdated.add(false);
			mChannelTopic.add("");
			mChannelNameMap.put(channel.toLowerCase(), num);
			mChannelUserList.add(new ArrayList<String>());
			final int channelId = mChannelName.size()-1;
			mChannelUserList.get(channelId).add(mNowNick);

			for (int i=listeners.length-1; i>=0; i--) {
				listeners[i].createChannel(mConfig.mServerName, channel, dateMsg);
			}

			// Talkのときは相手をユーザーリストに入れる
			final char c = channel.toCharArray()[0];
			if ((c != '%') && (c != '&') && (c != '#') && (c != '*') && (c != ' ')) {
				echoChannel(channel, "Start talking with " + channel, true);
				if (!mChannelUserList.get(channelId).contains(channel)) {
					mChannelUserList.get(channelId).add(channel);
				}
			} else {
				echoChannel(channel, "Channel " + IRCMsg.channelName(channel) + " has created", true);
			}
		}
	}

	/**
	 * チャンネルオブジェクトの削除
	 * @param channel チャンネル名
	 */
	public synchronized void removeChannel(final String channel) {
		if (V) Log.v(TAG, "removeChannel()");
		final String dateMsg = IRCMsg.getDateMsg();
		final int id = getChannelId(channel);
		if (id >= 0) {
			for (int i=listeners.length-1; i>=0; i--) {
				listeners[i].removeChannel(mConfig.mServerName, channel, dateMsg);
			}
			mSpanChannelLog.remove(id);
			mChannelName.remove(id);
			mChannelUpdated.remove(id);
			mChannelTopic.remove(id);
			mChannelNameMap.remove(channel.toLowerCase());
			mChannelUserList.remove(id);

			// カレントチャンネルの調整
			if ( getChannelId(mCurrentChannel) < 0 ) {
				final int newNum = mChannelName.size();
				final int newId = Math.min(id, newNum-1);
				if (newId < 0) {
					mCurrentChannel = null;
				} else {
					mCurrentChannel = mChannelName.get(newId);
				}
			}
		}
	}

	/**
	 * nickの変更
	 * @param oldNick 旧nick
	 * @param newNick 新nick
	 */
	public void changeNick(final String oldNick, final String newNick) {
		if (V) Log.v(TAG, "changeNick()");
		if (oldNick.equalsIgnoreCase(mNowNick)) {
			mNowNick = newNick;
		}
		for( int id=0; id<mChannelName.size(); id++) {
			final ArrayList<String> userList = mChannelUserList.get(id);
			final int uid = userList.indexOf(oldNick);
			if (uid >= 0) {
				userList.remove(uid);
			}
			if (!userList.contains(newNick)) {
				userList.add(newNick);
			}
		}
		final String dateMsg = IRCMsg.getDateMsg();
		for (int i=listeners.length-1; i>=0; i--) {
			listeners[i].changeNick(mConfig.mServerName, oldNick, newNick, dateMsg);
		}
	}

	/**
	 * メッセージの受信センター
	 * @param channel チャンネル名
	 * @param nick Nick
	 * @param message メッセージ
	 * @param isEcho エコーメッセージのときtrue
	 * @param update ログを更新するときtrue
	 * @param pale ログを淡色化するときtrue
	 */
	public void receiveMessageToChannel(final String channel, final String nick, final String message, final boolean isEcho, final boolean update, final boolean pale) {
		if (V) Log.v(TAG, "receiveMessageToChannel()");
		if (V) Log.v(TAG, "channel=[" + channel + "]");
		if (V) Log.v(TAG, "message=[" + message + "]");
		createChannel(channel);
		final int id = getChannelId(channel);
		if (id < 0) {
			if (D) throw new IllegalArgumentException("channel [" + channel + "] is not exist");
			return;
		}

		final String dateMsg = IRCMsg.getDateMsg();

		// キーワード確認
		boolean alert = false;
		int alertIndex = 0;
		int alertIndexEnd = 0;
		if (!pale) {
			if (mAlertKeywords != null) {
				final String capMsg = message.toUpperCase();
				for (int i=0; i<mAlertKeywords.length; i++) {
					alertIndex = capMsg.indexOf(mAlertKeywords[i]);
					if (alertIndex >= 0) {
						alert = true;
						alertIndexEnd = alertIndex + mAlertKeywords[i].length();
						break;
					}
				}
			}
		}

		// Spanログを作成する
		final SpannableStringBuilder ssb = new SpannableStringBuilder(message);
		if (pale) {
			final ForegroundColorSpan c = new ForegroundColorSpan(SystemConfig.paleLogColor);
			ssb.setSpan(c, 0, message.length(), Spanned.SPAN_EXCLUSIVE_EXCLUSIVE);
		} else {
			final ForegroundColorSpan c = new ForegroundColorSpan(SystemConfig.logColor);
			ssb.setSpan(c, 0, message.length(), Spanned.SPAN_EXCLUSIVE_EXCLUSIVE);
		}

		// リンクの作成
		Linkify.addLinks(ssb, Linkify.WEB_URLS);
		if (mConfig.mForTIG) {
			final Linkify.MatchFilter hashMatchFilter = new Linkify.MatchFilter() {
				public boolean acceptMatch(final CharSequence s, final int start, final int end) {
					if (start == 0) return true;
					final char c = s.charAt(start-1);
					if (IRCMsg.isUrlChar(c)) return false;
					return true;
				}
			};
			final Linkify.TransformFilter hashTransformFilter = new Linkify.TransformFilter() {
				public final String transformUrl(final Matcher match, final String url) {
					return match.group(1);
				}
			};
			final Pattern pattern = Pattern.compile("#([A-Za-z0-9_]+)");
			final String scheme = (SystemConfig.twitterSiteIsMobile ?  "http://mobile.twitter.com/searches?q=%23" : "http://twitter.com/search?q=%23");
			Linkify.addLinks(ssb, pattern, scheme, hashMatchFilter, hashTransformFilter);
		}

		if (alert) {
			final ForegroundColorSpan c = new ForegroundColorSpan(SystemConfig.alertKeywordsColor);
			ssb.setSpan(c, alertIndex, alertIndexEnd, Spanned.SPAN_EXCLUSIVE_EXCLUSIVE);
		}

		// spannableChannelLogへ収集
		ArrayList<SpannableStringBuilder> ssba = mSpanChannelLog.get(id);
		final int size = ssba.size();
		if ( size >= SystemConfig.spanChannelLogLines*2 ) {
			final ArrayList<SpannableStringBuilder> ssba2 = new ArrayList<SpannableStringBuilder>(
					ssba.subList(SystemConfig.spanChannelLogLines, size));
			ssba = ssba2;
		}

		final SpannableStringBuilder ssbLine = IRCMsg.colorDateMsg(dateMsg, SystemConfig.mainLogDateColor);
		if ((nick != null) && (nick.length() > 0)) {
			final SpannableStringBuilder ssbNick = IRCMsg.colorNick(nick);
			ssbLine.append(ssbNick);
		}
		ssbLine.append(ssb);
		ssbLine.append("\n");
		ssba.add(ssbLine);

		if (update) {
			mChannelUpdated.set(id, true);
		}

		// トークのときは常にalert
		if (!pale) {
			final char c = channel.toCharArray()[0];
			if ((c != '%') && (c != '&') && (c != '#') && (c != '*') && (c != ' ')) {
				alert = true;
			}
		}

		for (int i=listeners.length-1; i>=0; i--) {
			listeners[i].receiveMessageToChannel(mConfig.mServerName, channel, nick, dateMsg, ssb, alert);
		}
	}

	/**
	 * チャンネルへのメッセージ受信
	 * @param channel チャンネル名
	 * @param nick ニックネーム
	 * @param message メッセージ
	 * @param update ログを更新するときtrue
	 * @param pale ログを淡色化するときtrue
	 */
	public void receiveChannel(final String channel, final String nick, final String message, final boolean update, final boolean pale) {
		if (V) Log.v(TAG, "receiveChannel()");
		receiveMessageToChannel(
				channel,
				nick,
				message,
				false,
				update,
				pale
		);
	}

	/**
	 * システムチャンネルへのメッセージ受信
	 * @param message メッセージ
	 * @param pale ログを淡色化するときtrue
	 */
	public void receiveSystemChannel(final String message, final boolean pale) {
		if (V) Log.v(TAG, "receiveSystemChannel()");
		receiveMessageToChannel(
				IRCMsg.sSystemChannelName,
				null,
				message,
				false,
				false,
				pale
		);
	}

	//
	/**
	 * チャンネルへのメッセージエコー
	 * @param channel チャンネル名
	 * @param message メッセージ
	 * @param pale ログを淡色化するときtrue
	 */
	public void echoChannel(final String channel, final String message, final boolean pale) {
		if (V) Log.v(TAG, "echoChannel()");
		receiveMessageToChannel(
				channel,
				mNowNick,
				message,
				true,
				false,
				pale
		);
	}

	/**
	 * システムチャンネルへのメッセージエコー
	 * @param message メッセージ
	 */
	public void echoSystemChannel(final String message) {
		if (V) Log.v(TAG, "echoSystemChannel()");
		receiveMessageToChannel(
				IRCMsg.sSystemChannelName,
				null,
				message,
				true,
				false,
				true
		);
	}

	/**
	 * 指定チャンネルへのコマンドライン処理
	 * @param channel チャンネル名
	 * @param message コマンドライン
	 */
	public void sendCommandLine(final String channel, final String message) {
		if (V) Log.v(TAG, "sendCommandLine()");
		if (V) Log.v(TAG, "channel=[" + channel + "]");
		if (V) Log.v(TAG, "message=[" + message + "]");
		echoChannel(channel, message, true);
		final String words[] = message.split("[ ]+");
		final String cmd = words[0].toUpperCase();
		final String arg = new String(message).replaceAll("^[^ ]+[ ]+(.*)[ ]*", "$1");
		if (cmd.equals("/DEBUG")) {
			doDebug(channel, arg);
			return;
		}
		if (cmd.equals("/TOPIC")) {
			mConn.doTopic(channel, arg);
			return;
		}
		if (cmd.equals("/NICK")) {
			if (words.length >= 2) {
				mConn.doNick(words[1]);
			}
			return;
		}
		if (cmd.equals("/JOIN")) {
			if (words.length >= 2) {
				final char c = words[1].toCharArray()[0];
				if ((c != '%') && (c != '&') && (c != '#') && (c != '*') && (c != ' ')) {
					createChannel(words[1]);
				} else {
					mConn.doJoin(words[1]);
				}
			}
			return;
		}
		if (cmd.equals("/INVITE")) {
			if (words.length >= 2) {
				mConn.doInvite(words[1], channel);
			}
			return;
		}
		if (cmd.equals("/PART")) {
			mConn.doPart(channel, arg);
			return;
		}
		if (cmd.equals("/MODE")) {
			mConn.doMode(channel, arg);
			return;
		}
		if (cmd.equals("/KICK")) {
			final String arg2 = new String(message).replaceAll("^[^ ]+[ ]+[^ ]+[ ]+[^ ]+(.*)$", "$1");
			if (words.length >= 4) {
				mConn.doKick(channel, words[2], arg2);
			} else if (words.length >= 3) {
				mConn.doKick(channel, words[2], "kicked");
			}
			return;
		}
		if (cmd.equals("/WHOIS")) {
			if (words.length >= 2) {
				mConn.doWhois(words[1]);
			}
			return;
		}
		sendSystemChannel(message);
	}

	/**
	 * 指定チャンネルへのメッセージ送信
	 * @param channel チャンネル名
	 * @param message メッセージ
	 */
	public void sendChannel(final String channel, final String message) {
		if (V) Log.v(TAG, "sendChannel()");
		if (V) Log.v(TAG, "channel=[" + channel + "]");
		if (V) Log.v(TAG, "message=[" + message + "]");
		mConn.doPrivmsg(channel, message);
		echoChannel(channel, message, false);
	}

	/**
	 * システムチャンネルへのメッセージ送信
	 * @param message メッセージ
	 */
	public void sendSystemChannel(final String message) {
		if (V) Log.v(TAG, "sendSystemChannel()");
		if (V) Log.v(TAG, "message=[" + message + "]");
		mConn.send(message);
		//		echoChannel(IRCMsg.sSystemChannelName, message);
	}

	/**
	 * チャンネル数を得る（システムチャンネル含む）
	 * @return チャンネル数
	 */
	public int getNumChannels() {
		return mChannelName.size();
	}

	/**
	 * チャンネルIDからチャンネル名を得る
	 * @param id チャンネルID
	 * @return チャンネル名
	 */
	public String getChannelName(final int id) {
		if (V) Log.v(TAG, "getChannelName()");
		final int num=mChannelName.size();
		if (id < num) {
			return mChannelName.get(id);
		}
		return null;
	}

	/**
	 * チャンネル名リストを得る
	 * @return チャンネルリスト
	 */
	public String[] getChannelNameList() {
		if (V) Log.v(TAG, "getChannelList()");
		return mChannelName.toArray(new String[0]);
	}

	/**
	 * チャンネルログ更新リストを得る
	 * 配列順はgetChannelList()で得られるチャンネル名リストと同一
	 * @return チャンネルログ更新リスト
	 */
	public Boolean[] getChannelUpdatedList() {
		if (V) Log.v(TAG, "getChannelUpdated()");
		return mChannelUpdated.toArray(new Boolean[0]);
	}

	/**
	 * IDでSpannableチャンネルログを得る
	 * @param id チャンネルID
	 * @return Spannableチャンネルログ
	 */
	public SpannableStringBuilder getSpanChannelLog(final int id) {
		if (V) Log.v(TAG, "getChannelLog(" + id + ")");
		final int num=mChannelName.size();
		if (id < num) {
			final SpannableStringBuilder msg = new SpannableStringBuilder();
			final ArrayList<SpannableStringBuilder> ssba = mSpanChannelLog.get(id);
			for (int i=0; i<ssba.size(); i++) {
				msg.append(ssba.get(i));
			}
			return msg;

		}
		return null;
	}

	/**
	 * チャンネル名でSpannableチャンネルログを得る
	 * @param channel チャンネル名
	 * @return Spannableチャンネルログ
	 */
	public SpannableStringBuilder getSpanChannelLog(final String channel) {
		if (V) Log.v(TAG, "getChannelLog(" + channel + ")");
		final int id = getChannelId(channel);
		if (id < 0) {
			if (D) throw new IllegalArgumentException("channel [" + channel + "] is not exist");
			return null;
		}
		return getSpanChannelLog(id);
	}

	/**
	 * チャンネルIDでチャンネルログのクリア
	 * @param id チャンネルID
	 */
	public void clearChannelLog(final int id) {
		if (V) Log.v(TAG, "clearChannelLog(" + id + ")");
		final int num=mChannelName.size();
		if (id < num) {
			// mChannelLog.get(id).setLength(0);
			mSpanChannelLog.get(id).clear();
		}
	}

	/**
	 * チャンネル名でチャンネルログのクリア
	 * @param channel チャンネル名
	 */
	public void clearChannelLog(final String channel) {
		if (V) Log.v(TAG, "clearChannelLog(" + channel + ")");
		final int id = getChannelId(channel);
		if (id < 0) {
			if (D) throw new IllegalArgumentException("channel [" + channel + "] is not exist");
			return;
		}
		// mChannelLog.get(id).setLength(0);
		mSpanChannelLog.get(id).clear();
	}

	/**
	 * チャンネルIDでチャンネルログ更新フラグの取得
	 * @param id チャンネルID
	 * @return チャンネルログ更新フラグ
	 */
	public boolean getChannelUpdated(final int id) {
		if (V) Log.v(TAG, "clearChannelUpdated(" + id + ")");
		final int num=mChannelUpdated.size();
		if (id < num) {
			return mChannelUpdated.get(id);
		}
		return false;
	}

	/**
	 * チャンネル名でチャンネルログ更新フラグの取得
	 * @param channel チャンネル名
	 * @return チャンネルログ更新フラグ
	 */
	public boolean getChannelUpdated(final String channel) {
		if (V) Log.v(TAG, "clearChannelLog(" + channel + ")");
		final int id = getChannelId(channel);
		if (id < 0) {
			if (D) throw new IllegalArgumentException("channel [" + channel + "] is not exist");
			return false;
		}
		return mChannelUpdated.get(id);
	}

	/**
	 * チャンネルIDでチャンネルログ更新フラグのクリア
	 * @param id チャンネルID
	 */
	public void clearChannelUpdated(final int id) {
		if (V) Log.v(TAG, "clearChannelUpdated(" + id + ")");
		final int num=mChannelUpdated.size();
		if (id < num) {
			mChannelUpdated.set(id, false);
		}
	}

	/**
	 * チャンネル名でチャンネルログ更新フラグのクリア
	 * @param channel チャンネル名
	 */
	public void clearChannelUpdated(final String channel) {
		if (V) Log.v(TAG, "clearChannelLog(" + channel + ")");
		final int id = getChannelId(channel);
		if (id < 0) {
			if (D) throw new IllegalArgumentException("channel [" + channel + "] is not exist");
			return;
		}
		mChannelUpdated.set(id, false);
	}

	/**
	 * カレントチャンネルを得る
	 * @return チャンネル名
	 */
	public String getCurrentChannel() {
		if (V) Log.v(TAG, "getCurrentChannel()");
		return mCurrentChannel;
	}

	/**
	 * カレントチャンネルの設定
	 * @param channel チャンネル名
	 */
	public void setCurrentChannel(final String channel) {
		if (V) Log.v(TAG, "setCurrentChannel()");
		mCurrentChannel = new String(channel);
	}

	/**
	 * チャンネルに入る
	 * @param channel チャンネル名
	 */
	public void joinChannel(final String channel) {
		if (V) Log.v(TAG, "joinChannel()");
		mConn.doJoin(channel);
	}

	/**
	 * 複数チャンネルに順次入る
	 * @param channels チャンネル名（スペース区切り）
	 */
	public void joinChannels(final String channels) {
		if (V) Log.v(TAG, "joinChannels()");

		final Thread thread = new Thread(new Runnable() {
			public void run() {
				final String[] channelList = channels.split("[ \t\n]+");
				for ( int i=0; i<channelList.length; i++ ) {
					mConn.doJoin(channelList[i]);
					try {
						Thread.sleep(1000);
					} catch (final InterruptedException e) {
						//
					}
				}
			}
		});
		thread.start();
	}

	/**
	 * チャンネルから出る
	 * @param channel チャンネル名
	 */
	public void partChannel(final String channel) {
		if (V) Log.v(TAG, "joinChannel()");
		mConn.doPart(channel);
	}

	/**
	 * トピックを得る
	 * @param channel チャンネル名
	 * @return トピック
	 */
	public String getTopic(final String channel) {
		if (V) Log.v(TAG, "getTopic()");
		final int id = getChannelId(channel);
		if (id < 0) {
			if (D) throw new IllegalArgumentException("channel [" + channel + "] is not exist");
			return null;
		}
		return mChannelTopic.get(id).toString();
	}

	/**
	 * トピックを変更する
	 * @param channel チャンネル名
	 * @param topic トピック
	 */
	public void changeTopic(final String channel, final String topic) {
		if (V) Log.v(TAG, "setTopic()");
		createChannel(channel);
		final int id = getChannelId(channel);
		if (id < 0) {
			if (D) throw new IllegalArgumentException("channel [" + channel + "] is not exist");
			return;
		}
		mChannelTopic.set(id, topic);
		final String dateMsg = IRCMsg.getDateMsg();
		for (int i=listeners.length-1; i>=0; i--) {
			listeners[i].changeTopic(mConfig.mServerName, channel, topic, dateMsg);
		}
	}

	/**
	 * nickを得る
	 * @return nick
	 */
	public String getNick() {
		if (V) Log.v(TAG, "getNick()");
		return mConn.getNick();
	}

	/**
	 * nickの設定
	 * @param nick nick
	 */
	public void setNick(final String nick) {
		if (V) Log.v(TAG, "setNick()");
		mConn.doNick(nick);
	}

	/**
	 * 接続中かどうかを返す
	 * @return 接続中ならtrue
	 */
	public boolean isConnected() {
		if (V) Log.v(TAG, "isConnected()");
		return (mConn != null) && mConn.isConnected();
	}

	/**
	 * 再接続時刻読み出し
	 * @return 再接続時刻
	 */
	public long getNextReconnectTime() {
		if (V) Log.v(TAG, "getNextReconnectTime()");
		return mNextReconnectTime;
	}

	/**
	 * 再接続時刻登録
	 */
	public void setNextReconnectTime() {
		if (V) Log.v(TAG, "setNextReconnectTime()");
		final int t1 = SystemConfig.autoReconnectWaitSec;
		if (t1 > 0) {
			final long count = getTimerCount() + t1;
			mNextReconnectTime = count;
		}
	}

	/**
	 * 再接続時刻クリア
	 */
	public void clearNextReconnectTime() {
		if (V) Log.v(TAG, "clearNextReconnectTime()");
		mNextReconnectTime = 0;
	}

	/**
	 * KeepAlive送出時刻読み出し
	 * @return 送出時刻
	 */
	public long getNextKeepAliveTime() {
		if (V) Log.v(TAG, "getNextKeepAliveTime()");
		return mNextKeepAliveTime;
	}

	/**
	 * KeepAlive送出時刻登録
	 */
	public void setNextKeepAliveTime() {
		if (V) Log.v(TAG, "setNextKeepAliveTime()");
		if (isConnected()) {
			final int t1 = SystemConfig.checkConnectWaitSec;
			if (t1 > 0) {
				final long count = getTimerCount() + t1;
				mNextKeepAliveTime = count;
			}
		}
	}

	/**
	 * KeepAlive送出停止
	 */
	public void clearNextKeepAliveTime() {
		if (V) Log.v(TAG, "clearNextKeepAliveTime()");
		mNextKeepAliveTime = 0;
	}

	/**
	 * KeepAliveデータを送出する
	 */
	public void sendKeepAliveMessage() {
		if (V) Log.v(TAG, "sendKeepAliveMessage()");
		if (mNowNick != null) {
			if (isConnected()) {
				if (V) Log.v(TAG, "send keepalive message");
				if (SystemConfig.verbose >= 2) echoSystemChannel("send keepalive message");
				mConn.doPong(mNowNick);
			}
		}
	}

	/**
	 * 自動再接続フラグ読み出し
	 * @return 自動再接続フラグがonならtrue
	 */
	public boolean getNeedReconnect() {
		if (V) Log.v(TAG, "getNeedReconnect()");
		if (!mNeedReconnect) {
			if (V) Log.v(TAG, "not need reconnect");
			if (SystemConfig.verbose >= 2) echoSystemChannel("not need reconnect");
			return false;
		}
		return mConfig.mAutoReconnect;
	}

	/**
	 * 接続が切れているかどうか調べる
	 * @return 接続が切れているならtrue
	 */
	public boolean isDisconnected() {
		if (V) Log.v(TAG, "isDisconnected()");
		if (!mNeedReconnect) {
			if (V) Log.v(TAG, "not need reconnect");
			if (SystemConfig.verbose >= 2) echoSystemChannel("not need reconnect");
			return false;
		}
		return !isConnected();
	}

	/**
	 * ユーザーリストを得る
	 * @param channel チャンネル名
	 * @return ユーザーリスト
	 */
	public String[] getUserList(final String channel) {
		if (V) Log.v(TAG, "getUserList()");
		final int id = getChannelId(channel);
		if (id < 0) {
			if (D) throw new IllegalArgumentException("channel [" + channel + "] is not exist");
			return null;
		}
		final ArrayList<String> userList = mChannelUserList.get(id);
		final String[] users = userList.toArray(new String[0]);
		java.util.Arrays.sort(users);
		return users;
	}

	/**
	 * サーバ設定がTIGモードかどうか調べる
	 * @return サーバ設定がTIGモードならtrue
	 */
	public boolean isTIGMode() {
		if (V) Log.v(TAG, "isTIGMode()");
		return mConfig.mForTIG;
	}

	/**
	 * IRClibリスナ
	 */
	private class IRClibListener implements IRCEventListener{

		public void onRegistered() {
			receiveSystemChannel("Connected " + mConfig.mServerName, true);
			for (int i=listeners.length-1; i>=0; i--) {
				listeners[i].receiveConnect(mConfig.mServerName);
			}
			joinChannels(mConfig.mChannel);
			setNextKeepAliveTime();
			clearNextReconnectTime();
		}

		public void onDisconnected() {
			receiveSystemChannel("Disconnected " + mConfig.mServerName, true);
			for (int i=listeners.length-1; i>=0; i--) {
				listeners[i].receiveDisconnect(mConfig.mServerName);
			}
			clearNextKeepAliveTime();
			if (getNeedReconnect()) {
				setNextReconnectTime();
			}
		}

		public void onError(final String message) {
			receiveSystemChannel("IRC Error: " + message, true);
			setNextKeepAliveTime();
		}

		public void onError(final int num, final String message) {
			// receiveSystemChannel("IRC Error: " + num + " " + message);
			if (num > 0) {
				doNumericReply(num, "", message);
			} else {
				receiveSystemChannel("IRC Error: " + num + " " + message, true);
			}
			setNextKeepAliveTime();
		}

		public void onInvite(final String channel, final IRCUser user, final String invitedNick) {
			joinChannels(channel);
			receiveChannel(channel, null, user.getNick() + " has invited " + invitedNick + " to " + channel, false, true);
			setNextKeepAliveTime();
		}

		public void onJoin(final String channel, final IRCUser user) {
			final int channelId = getChannelId(channel);
			if (channelId >= 0) {
				if (!mChannelUserList.get(channelId).contains(user.getNick())) {
					mChannelUserList.get(channelId).add(user.getNick());
				}
			}
			receiveChannel(channel, null, user.getNick() + " has joined " + channel, false, true);
			setNextKeepAliveTime();
		}

		public void onKick(final String channel, final IRCUser user, final String kickedNick, final String message) {
			receiveChannel(channel, null, user.getNick() + " has kicked " + kickedNick, false, true);
			final int channelId = getChannelId(channel);
			if (channelId >= 0) {
				if (mChannelUserList.get(channelId).contains(user.getNick())) {
					mChannelUserList.get(channelId).remove(user.getNick());
				}
			}
			setNextKeepAliveTime();
		}

		public void onMode(final String channel, final IRCUser user, final IRCModeParser modeParser) {
			receiveChannel(channel, null, user.getNick() + " has set mode " + modeParser.getLine(), false, true);
			setNextKeepAliveTime();
		}

		public void onMode(final IRCUser user, final String targetNick, final String mode) {
			receiveSystemChannel(user.getNick() + " has changed the mode of " + targetNick + " to " + mode, true);
			setNextKeepAliveTime();
		}

		public void onNick(final IRCUser user, final String newNick) {
			receiveSystemChannel(user.getNick() + " is now known as " + newNick, true);
			changeNick(user.getNick(), newNick);
			setNextKeepAliveTime();
		}

		public void onNotice(final String target, final IRCUser user, final String message) {
			if (user.getNick() == null) {
				receiveSystemChannel(IRCMsg.channelNickMessage(target, user.getNick(), message), true);
				// } else if (user.getHost() == null) {
				// 	receiveSystemChannel(IRCMsg.channelNickMessage(target, user.getNick(), message), true);
			} else {
				final char c = target.toCharArray()[0];
				if ((c != '%') && (c != '&') && (c != '#') && (c != '*') && (c != ' ')) {
					receiveChannel(user.getNick(), user.getNick(), message, true, true);
				} else {
					receiveChannel(target, user.getNick(), message, true, true);
				}
			}
			setNextKeepAliveTime();
		}

		public void onPart(final String channel, final IRCUser user, final String message) {
			receiveSystemChannel(user.getNick() + " has left channel (" + message + ")", true);
			final int channelId = getChannelId(channel);
			if (channelId >= 0) {
				if (mChannelUserList.get(channelId).contains(user.getNick())) {
					mChannelUserList.get(channelId).remove(user.getNick());
				}
			}
			setNextKeepAliveTime();
		}

		public void onPrivmsg(final String channel, final IRCUser user, final String message) {
			if (user.getNick() == null) {
				receiveSystemChannel(IRCMsg.channelNickMessage(channel, user.getNick(), message), true);
			} else if (user.getHost() == null) {
				receiveSystemChannel(IRCMsg.channelNickMessage(channel, user.getNick(), message), true);
			} else {
				final char c = channel.toCharArray()[0];
				if ((c != '%') && (c != '&') && (c != '#') && (c != '*') && (c != ' ')) {
					receiveChannel(user.getNick(), user.getNick(), message, true, false);
				} else {
					receiveChannel(channel, user.getNick(), message, true, false);
				}
			}
			setNextKeepAliveTime();
		}

		public void onQuit(final IRCUser user, final String message) {
			receiveSystemChannel(user.getNick() + " has left IRC (" + message + ")", true);
			setNextKeepAliveTime();
		}

		public void onReply(final int num, final String value, final String message) {
			if (num > 0) {
				doNumericReply(num, value, message);
			} else {
				receiveSystemChannel("[" + Integer.toString(num) + "] [" + value + "] ["+ message + "]", true);
			}
			setNextKeepAliveTime();
		}

		public void onTopic(final String channel, final IRCUser user, final String topic) {
			changeTopic(channel, topic);
			receiveChannel(channel, "", user.getNick() + " has changed the topic to " + topic, false, true);
			setNextKeepAliveTime();
		}

		public void onPing(final String ping) {
			if (SystemConfig.verbose >= 1) {
				receiveSystemChannel("Ping(): doPong " + ping, true );
			}
			mConn.doPong(ping);
			setNextKeepAliveTime();
		}

		public void unknown(final String prefix, final String command, final String middle, final String trailing) {
			receiveSystemChannel(command + " " + middle + " :" + trailing, true);
			// receiveSystemChannel("unknown(): prefix=[" + prefix + "] command=[" + command + "] middle=[" + middle + "] trailing=[" + trailing + "]");
			setNextKeepAliveTime();
		}

	}

	/**
	 * SSLトラストマネージャ
	 */
	public class TrustManager implements SSLTrustManager {
		private X509Certificate[] mChain;
		public X509Certificate[] getAcceptedIssuers() {
			return mChain != null ? mChain : new X509Certificate[0];
		}
		public boolean isTrusted(final X509Certificate[] chain) {
			mChain = chain;
			return true;
		}
	}

	/**
	 * ニューメリックリプライの処理
	 * @param num ニューメリックリプライ番号
	 * @param value ?
	 * @param message メッセージ
	 */
	public void doNumericReply(final int num, final String value, final String message) {
		// receiveSystemChannel("[" + Integer.toString(num) + "] [" + value + "] ["+ message + "]");
		switch (num) {
			case 332:  // RPL_TOPIC
			{
				final String[] words = value.split("[ ]+");
				if (words.length >= 2) {

					changeTopic(words[1], message);
				}
				break;
			}

			case 353:  // RPL_NAMREPLY
			{
				final String[] words = value.split("[ ]+");
				final String channel;
				if (words.length < 3) break;
				channel = words[2];
				final int channelId = getChannelId(channel);
				if (channelId < 0) break;
				final String[] users = message.split("[ ]+");
				for (int i=0; i<users.length; i++) {
					final String user = users[i].replaceAll("^[@+]", "");
					if (!mChannelUserList.get(channelId).contains(user)) {
						mChannelUserList.get(channelId).add(user);
					}
				}
				// receiveSystemChannel("[" + Integer.toString(num) + "] [" + value + "] ["+ message + "]");
				break;
			}
			case 366:  // RPL_ENDOFNAMES
				break;

			case 372:  // RPL_MOTD
			case 375:  // RPL_MOTDSTART
				receiveSystemChannel(message, true);
				break;
			case 376:  // RPL_ENDOFMOTD
			case 422:  // RPL_ERR_NOMOTD
				break;

			case   1:  //
			case   2:  //
			case   3:  //
			case   4:  //
			case   5:  //
				receiveSystemChannel(message, true);
				break;

			case 433:  // RPL_NICKNAMEISUSE
			{
				// receiveSystemChannel("[" + Integer.toString(num) + "] [" + value + "] ["+ message + "]");
				// NICK衝突回避
				final int len = mNowNick.length();
				final char[] nickChr = mNowNick.toCharArray();
				int i;
				for (i=len; i>0; --i) {
					if ((nickChr[i-1] < '0') || (nickChr[i-1] > '9')) break;
				}
				int suffix = 0;
				for (int j=i; j<len; j++) {
					suffix = suffix*10 + (nickChr[j]-'0');
				}
				final String newSuffix = Integer.toString(suffix+1);
				int nickRestLen = len-newSuffix.length();
				if (nickRestLen < 0) nickRestLen = 0;
				final String newNick = mNowNick.substring(0, nickRestLen) + newSuffix;
				echoSystemChannel("newNick is [" + newNick + "]");
				changeNick(mConn.getNick(), newNick);
				break;
			}

			case 311:  // RPL_WHOISUSER
			{
				final String[] words = value.split("[ ]+");
				if (words.length >= 2) {
					echoSystemChannel("Nick: [" + words[1] + "]");
				}
				if (words.length >= 3) {
					echoSystemChannel("User Name: [" + words[2] + "]");
				}
				if (words.length >= 4) {
					echoSystemChannel("Host: [" + words[3] + "]");
				}
				echoSystemChannel("Real Name: [" + message + "]");
				break;
			}

			case 319:  // RPL_WHOISCHANNELS
			{
				final String[] words = message.split("[ ]+");
				for (int i=0; i<words.length; i++) {
					echoSystemChannel("Channel: [" + words[i] + "]");
				}
				break;
			}

			case 312:  // RPL_WHOISSERVER
			{
				final String[] words = value.split("[ ]");
				if (words.length >= 3) {
					echoSystemChannel("Server: [" + words[2] + "]");
				}
				echoSystemChannel("Server Information: [" + message + "]");
				break;
			}

			case 251:  // RPL_LUSERCLIENT
			case 313:  // RPL_WHOISOPERATOR
			case 317:  // RPL_WHOISIDLE
			case 318:  // RPL_ENDOFWHOIS
			case 252:  //
			case 253:  //
			case 254:  // RPL_LUSERCHANNELS
			case 255:  // RPL_LUSERME
			case 333:  //
			case 352:  // WHOREPLY
			case 315:  // ENDOFWHO
			default:
				if (SystemConfig.verbose >= 1) {
					receiveSystemChannel("[" + Integer.toString(num) + "] [" + value + "] ["+ message + "]", true);
				}
				break;
		}
	}

	/**
	 * ランダム文字列を生成する。
	 * @param length 文字列長
	 * @return 文字列
	 */
	public String randomMessage(final int length) {
		final String chars = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
		final StringBuffer str = new StringBuffer(length+1);
		for (int i=0; i<length; i++) {
			final int idx = (int)(Math.random()*chars.length());
			str.append(chars.substring(idx, idx+1));
		}
		return str.toString();
	}

	/**
	 * デバッグコマンド処理
	 * @param channel チャンネル
	 * @param line デバッグコマンドライン
	 */
	public void doDebug(final String channel, final String line) {
		if (V) Log.v(TAG, "doDebug()");
		final String words[] = line.split(" ");
		final String cmd = words[0].toUpperCase();
		final String arg = new String(line).replaceAll("^[^ ]+[ ]+(.*)[ ]*", "$1");
		if (cmd.equals("MSG")) {
			receiveChannel(channel, mNowNick, arg, true, true);
			return;
		}
		if (cmd.equals("REPLY")) {
			final int numericReply = Integer.valueOf(words[1]);
			doNumericReply(numericReply, mNowNick, arg);
			return;
		}
		if (cmd.equals("RNDMSG")) {
			int count = 100;
			if (words.length >= 2) {
				count = Integer.valueOf(words[1]);
			}
			for (int i=0; i<count; i++) {
				final String message = randomMessage(64);
				receiveChannel(channel, mNowNick, message, true, true);
			}
			return;
		}
		if (cmd.equals("RNDCHMSG")) {
			int count = 100;
			if (words.length >= 2) {
				count = Integer.valueOf(words[1]);
			}
			for (int i=0; i<count; i++) {
				final String message = randomMessage(64);
				receiveChannel(mChannelName.get((int)(Math.random()*mChannelName.size())), mNowNick, message, true, true);
			}
			return;
		}
	}

	/**
	 * タイマー処理
	 */
	private boolean mTimerStop = false;
	private static final long mTimerStep = 1000;
	private long mTimerCount = 0;
	private long mTimerNext = 0;
	private static final int TIMER = 1;
	private final Handler mTimerHandler = new Handler() {
		@Override
		public void handleMessage(final Message msg) {
			if (msg.what == TIMER) {
				while (SystemClock.uptimeMillis() > mTimerNext) {
					mTimerNext += mTimerStep;
					mTimerCount++;
					onTimer(mTimerCount);
				}
				if (mTimerStop == false) {
					final Message msg2 = mTimerHandler.obtainMessage(TIMER);
					sendMessageAtTime(msg2, mTimerNext);
				}
			}
		}
	};
	private void startTimer() {
		if (V) Log.v(TAG, "startTimer()");
		mTimerStop = false;
		mTimerNext = SystemClock.uptimeMillis();
		final Message msg = mTimerHandler.obtainMessage(TIMER);
		mTimerHandler.sendMessageAtTime(msg, mTimerNext+mTimerStep);
	}
	private void stopTimer() {
		if (V) Log.v(TAG, "stopTimer()");
		mTimerStop = true;
	}
	private long getTimerCount() {
		return mTimerCount;
	}
	private void onTimer(final long timerCount) {
		// KeepAlive送出＆切断検出
		final long t1 = getNextKeepAliveTime();
		if ((t1 > 0) && (t1 < timerCount)) {
			if (SystemConfig.checkConnectWaitSec > 0) {
				if (D) Log.d(TAG, "ontimer(): try keepalive");
				if (SystemConfig.verbose >= 1) echoChannel(IRCMsg.sSystemChannelName, "Post KeepAlive packet", true);
				setNextKeepAliveTime();
				sendKeepAliveMessage();
				if (isDisconnected()) {
					clearNextKeepAliveTime();
					if (D) Log.d(TAG, "ontimer(): lost connection");
					echoChannel(IRCMsg.sSystemChannelName, "Lost connection", true);
					disconnect();
					connect();
				}
			} else {
				clearNextKeepAliveTime();
			}
		}

		// 自動再接続
		final long t2 = getNextReconnectTime();
		if ((t2 > 0) && (t2 < timerCount)) {
			if (D) Log.d(TAG, "ontimer(): try reconnect");
			echoChannel(IRCMsg.sSystemChannelName, "Try reconnect", true);
			clearNextReconnectTime();
			connect();
		}

	}



}

